package uk.co.hcsoftware.yago;

/**
 * Extension to Yago option, that allows setting of a default option value,
 * which will be returned as the arg value if it has not been set explicitly
 * on command line.
 * 
 * The default value can also be queried directly.
 *
 *
 * @author DM
 */
public class DefaultedOption<V> extends Option<V> {

  V defaultValue;

  /**
   * Corresponds to similar <code>Option</code> constructor, but here with a default 
   * value replacing the <code>required</code> parameter (if a default value set,
   * then parameter implicitly not required on command line).
   * 
   * @param defaultValue The value <code>getArgValue()</code> will return if the argument
   * is not set on the command line.
   * 
   * @see uk.co.hcsoftware.yago.Option#Option(java.lang.String, boolean, uk.co.hcsoftware.yago.ArgReader, java.lang.String)
   */
  public DefaultedOption(String name, V defaultValue, ArgReader<V> argReader, String description) {
    super(name, false, argReader, description);
    this.defaultValue = defaultValue;
  }

  /**
   * Corresponds to similar <code>Option</code> constructor, but here with a default 
   * value replacing the <code>required</code> parameter (if a default value set,
   * then parameter implicitly not required on command line).
   * 
   * @param defaultValue The value <code>getArgValue()</code> will return if the argument
   * is not set on the command line.
   *
   * @see uk.co.hcsoftware.yago.Option#Option(java.lang.String, boolean, char, java.lang.String, uk.co.hcsoftware.yago.ArgReader, java.lang.String)
   */
  public DefaultedOption(String name, V defaultValue, char shortId, String longId, ArgReader<V> argReader, String description) {
    super(name, false, shortId, longId, argReader, description);
    this.defaultValue = defaultValue;
  }

  /**
   * Corresponds to similar <code>Option</code> constructor, but here with a default 
   * value replacing the <code>required</code> parameter (if a default value set,
   * then parameter implicitly not required on command line).
   * 
   * @param defaultValue The value <code>getArgValue()</code> will return if the argument
   * is not set on the command line.
   * 
   * @see uk.co.hcsoftware.yago.Option#Option(java.lang.String, boolean, java.lang.String)
   */
  public DefaultedOption(String name, V defaultValue, String description) {
    super(name, false, description);
    this.defaultValue = defaultValue;
  }

  /**
   * @return The default value that will be returned by <code>getArgValue()</code> if
   *   not explicitly set on command line.
   */
  public V getDefaultArgValue() {
    return defaultValue;
  }

  /**
   * If a default value has been set, but no argument value parsed from command line,
   * then returns 
   * 
   * @see uk.co.hcsoftware.yago.Option#getArgValue()
   */
  @Override
  public V getArgValue() {
    if (super.getArgValue() == null) {
      return defaultValue;
    }
    return super.getArgValue();
  }
}
