package uk.co.hcsoftware.yago;

import uk.co.hcsoftware.yago.parsers.GnuParser;
import uk.co.hcsoftware.yago.parsers.IncompleteGroupException;
import uk.co.hcsoftware.yago.parsers.InvalidGroupException;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * argument parser. Use something like:
 * o=new Option(..);
 * opts.add(o);
 * a=new Arg(..);
 * args.add(a);
 * p=new Parser(opts,args);
 * p.parse(); //may throw ParseException of various kinds
 * o.getArgValue()
 * a.getArgValue()
 *
 * the weird thing is that you use the same objects you passed to the parser to retrieve the values
 * This actually makes calling code much terser than the usual argparsing libs as there is just 1 place
 * where each yago arg is defined
 *
 * @see uk.co.hcsoftware.yago.util.DefaultArgReaders for some helper classes
 *
 *
 */

public abstract class Parser {
    public static final Logger log = Logger.getLogger(Parser.class.getName());
    protected final List<Option> options;
    protected final List<Argument> args;
    protected List<String> extraArgs=new ArrayList<String>();
    private List<List<Option>> atLeastOneOptionGroups = new ArrayList<List<Option>>();
    private List<List<Option>> exclusiveOptionGroups = new ArrayList<List<Option>>();
    private List<List<Option>> bundleOptionGroups = new ArrayList<List<Option>>();

    public Parser(Option ... options){
        this(options,null);
    }

    public Parser(Option[] options, Argument[] args){
        this(options==null?Collections.EMPTY_LIST:Arrays.asList(options)
                ,args==null?Collections.EMPTY_LIST:Arrays.asList(args)
        );
    }
    public Parser(List<Option> options, List<Argument> args){
        this.options = options==null?Collections.EMPTY_LIST:options;
        this.args = args==null?Collections.EMPTY_LIST:args;
    }

    public void parse(String[] argv) throws ParseException{

        parseArgs(argv);
        //check for required
        checkRequiredElementsPresent();
        checkOptionGroups();
    }


    protected abstract void parseArgs(String[] argv) throws ParseException;


    /**
     * parses the options, and sets their values or prints errmessage and calls System.exit(1)
     * @deprecated for back-compatibility, use parseOrDie(String,String,String[]) instead
     */
    public void parseOrDie(String programInvocation, String[] args){
        parseOrDie(programInvocation,args,null);
    }
    /**
     *
     * @param programInvocation
     * @param args
     * @param longDescription
     * @deprecated for back-compatibility, use parseOrDie(String,String,String[]) instead
     */
    public void parseOrDie(String programInvocation, String[] args, String longDescription){
        StringBuilder sb = new StringBuilder();
        sb.append("usage: ");
        sb.append(programInvocation);
        boolean allOptionsOptional=true;
        if(!options.isEmpty()){
            for(Option o:options) {
                if(o.isRequired()) {
                    allOptionsOptional=false;
                    break;
                }
            }
            if (allOptionsOptional) {
                sb.append(" [options] ");
            } else {
                sb.append(" options ");
            }
        }
        if (!this.args.isEmpty()) {
            sb.append(" ");
            sb.append(getArgsSummaryString());
        }

        //now the long description if there is one
        if(longDescription!=null && longDescription.length()>0){
            sb.append('\n');
            sb.append(longDescription);
        }

        //now for each arg have the descrip
        if(!this.args.isEmpty()){
            sb.append('\n');
            for(Argument a:this.args){
                final String description = a.getDescription();
                if (description!=null && description.length()>0) {
                    sb.append(a.getName());
                    sb.append(" ");
                    sb.append(description);
                }
            }
        }

        parseOrDie(sb.toString(), "", args);

    }

    public void parseOrDie(String usageHeader, String usageFooter, String[] args){
        try {
            parse(args);
            return;
        } catch (ParseException e) {
            log.log(Level.FINE, "ParseException", e);
            System.out.println(e.getFriendlyMessage());
            System.out.println("(invocation: "+argsToString(args)+")");
            System.out.println(getUsage(usageHeader, usageFooter));
        }
        System.exit(1);
    }


    private String argsToString(String[] args) {
        StringBuilder sb = new StringBuilder();
            for(String f:args){
                if(sb.length()>0) sb.append(' ');
                sb.append(f);
            }
            return sb.toString();
    }

    public List<String> getExtraArgs() {
        return extraArgs;
    }


    //reading args is pretty much the same for all parsers, so keep the code here where they can share it
    //override if desired, but should keep behaviour similar.. (i.e. add to extraargs and call Argument.read(..)
    protected void readArg(int argIndex, String s) throws InvalidArgException {
        if(argIndex>-1 && argIndex<args.size()){
            log.fine("read argIndex="+argIndex+" val="+s);
            args.get(argIndex).read(s, argIndex);
        }
        else{
            log.fine("extraarg argIndex="+argIndex+" val="+s);
            extraArgs.add(s);
        }
    }

    protected void checkRequiredElementsPresent() throws MissingOptException, MissingArgException {
        for(Option o:options){
            if(o.isRequired() && !o.isSet()){
                throw new MissingOptException(o);
            }
        }
        int argIndex=0;
        for(Argument o:args){
            if(o.isRequired() && !o.wasFound()){
                throw new MissingArgException(argIndex, o);
            }
            argIndex++;
        }
    }


    public abstract String getUsage(String usageHeader, String usageFooter);

    protected void pad(StringBuilder sb, int charcount){
        char[] c = new char[charcount];
        Arrays.fill(c,' ');
        sb.append(c);
    }

    public String getLeftJustifiedText(int firstLineIndent, int lmargin,int rmargin,String text){
        StringBuilder sb = new StringBuilder();
        pad(sb,firstLineIndent);
        int w=rmargin-lmargin-1; //allow 1 char for carriage returns we insert
        for(int i=0;i<text.length();){
            sb.append(text.subSequence(i,Math.min(text.length(),i+w)));
            i+=w;
            if(i<text.length()){
                sb.append('\n');
                pad(sb,lmargin);
                //now eat any whitespace @ start of line
                while(i<text.length() && Character.isWhitespace(text.charAt(i))){
                    i++;
                }
            }
        }
        return sb.toString();
    }

    public List<Option> getSetOptions() {
        List<Option> opts = new ArrayList<Option>(options.size());
        for(Option o:options){
            if(o.isSet()) opts.add(o);
        }
        return opts;
    }
    public List<Argument> getFoundArguments() {
        List<Argument> opts = new ArrayList<Argument>(options.size());
        for(Argument o:args){
            if(o.wasFound()) opts.add(o);
        }
        return opts;
    }
    /**
     * List<Object> : extraArgs are all String defined args can be any type
     * @return list of all defined args, then extra args
     */
    public List getAllArguments(){
        List<Argument> f=getFoundArguments();
        List l = new ArrayList(f.size()+extraArgs.size());
        for(Argument a:f) l.add(a.getArgValue());
        l.addAll(extraArgs);
        return l;
    }
    public int getTotalArgCount(){
        return args.size()+extraArgs.size();
    }
         //todo: move the printout stuff into a separate class (think it's OK to have it invoked by the parser, as
    //diff parsers have diff usage printouts
    public String getOptionsString(int lmargin,int width) {
        //sort alphabetically
        List<Option> options = new ArrayList<Option>(this.options);
        Collections.sort(options, new Comparator<Option>() {
            public int compare(Option option, Option option1) {
                return option.getName().compareTo(option1.getName());
            }
        });
        StringBuilder sb = new StringBuilder();
        int maxOptWidth=0;
        for(Option o:options){
            int mw = getOptSummary(o).length();
            if(mw>maxOptWidth) maxOptWidth=mw;
        }
        int rmargin=width-lmargin;
        final int gutter = 2;
        int tab=maxOptWidth+gutter+lmargin;
        if(tab>rmargin/2) tab=rmargin/2;

        for(Option o:options){
            int cursor=0;
            pad(sb,lmargin);
            cursor+=lmargin;
            final String optSummary = getOptSummary(o);
            sb.append(optSummary);
            cursor+=optSummary.length();
            if (cursor>rmargin) {
                cursor=rmargin%cursor;
            }
            int spaceToTab=tab-cursor;
            GnuParser.log.fine("spacetotab="+spaceToTab);
            if(spaceToTab<1){//start text on next line
                sb.append('\n');
                pad(sb,tab);
            }
            else{
                pad(sb,spaceToTab);
            }
            String d = o.getDescription();
            if(d!=null){
                //wrap d to rmargin-tab chars
                sb.append(getLeftJustifiedText(0,tab+1,rmargin,d));
            }
            sb.append('\n');
        }
        return sb.toString();
    }

    public String getOptSummary(Option o) {
        StringBuilder sb = new StringBuilder();
        if(o.hasShortId()){
            sb.append('-');
            sb.append(o.getShortId());
        }
        if(o.hasLongId()){
            if(sb.length()>0) sb.append(',');
            sb.append("--");
            sb.append(o.getLongId());
        }
        if(o.takesArgument()){
            if(o.getArgName()!=null){
                sb.append(" <"+o.getArgName()+">");
            }
            else sb.append(" <arg>");
        }
        if(o.isRequired()) sb.append(" *");
        return sb.toString();
    }

    public String getArgsSummaryString() {
        StringBuilder sb = new StringBuilder();
        int i=0;
        for(Argument a: args){
            sb.append(getArgString(a,"arg"+i));
            i++;
        }
        return sb.toString();
    }

    private String getArgString(Argument a, String defaultName) {
        StringBuilder sb = new StringBuilder();
        if(!a.isRequired()) sb.append('[');
        if(a.getName()==null){
            sb.append(defaultName);
        }
        else sb.append(a.getName());
        if(!a.isRequired()) sb.append(']');
        return sb.toString();
    }



    /**
     * >=1 of these must be present
     * @param opts
     */
    public void createAtLeastOneOptionGroup(Option... opts){
        atLeastOneOptionGroups.add(Arrays.asList(opts));
    }

    /**
     * no more than one of these can be present
     * @param opts
     */
    public void createExclusiveOptionGroup(Option... opts){
        exclusiveOptionGroups.add(Arrays.asList(opts));
    }

    /**
     * if any of these is present they all must be
     * @param opts
     */
    public void createBundleOptionGroup(Option... opts){
        bundleOptionGroups.add(Arrays.asList(opts));
    }

    private void checkOptionGroups() throws IncompleteGroupException, InvalidGroupException {
        Set opts = new HashSet(getSetOptions());
        //check bundles
        for(List<Option> bundle:bundleOptionGroups){
            if(containsAny(opts,bundle) && !opts.containsAll(bundle)){
                throw new IncompleteGroupException(bundle);
            }
        }
        for(List<Option> bundle:exclusiveOptionGroups){
            if(containsMoreThanOne(opts,bundle)){
                throw new InvalidGroupException("only one of these allowed",bundle);
            }
        }
        for(List<Option> bundle:atLeastOneOptionGroups){
            if(!containsAny(opts,bundle)){
                throw new InvalidGroupException("at least one of these required",bundle);
            }
        }
    }

    private boolean containsMoreThanOne(Set s, Collection subset){
        int count=0;
        for(Object o: subset){
            if(s.contains(o)) count++;
            if(count>1) return true;
        }
        return false;
    }

    private boolean containsAny(Set s, Collection subset){
        for(Object o: subset){
            if(s.contains(o)) return true;
        }
        return false;
    }
}
