package uk.co.hcsoftware.yago.parsers;

import uk.co.hcsoftware.yago.*;

import java.util.*;
import java.util.logging.Logger;



/**
 * <p> parser with similar behaviour to GNU getopt: http://www.urbanophile.com/arenn/hacking/getopt/gnu.getopt.Getopt.html</p>
 *
 *<p> from what I can make out, reading the 2000wd essay in the javadocs:</p>
 *  <ul>
 *<li> options must start - or --  (everything else is an arg) (unless in posixlyCorrect mode)</li>
 *<li>  short options start -  and have 1char identifier </li>
 * <li> long options start --  and have string identifier</li>
 *<li>  -- on its own means everything following (regardless of above) is an arg </li>
 * <li>  ordering of options/arguments doesn't matter (except wrt previous point)</li>
 *<li>  -axf is equivalent to -a -x -f </li>
 *<li>  --longopt=a is equivalent to --longopt a </li>
 *<li>  can abbreviate long options to as few chars as will uniquely identify it. Failure to uniquely identify causes exception</li>
 * <li> if something starts with - or -- and doesn't match some option causes UnrecognizedOptionException
 *  (unless in posixlyCorrect mode)
 * </li>
 * <li> -W is special: -W outputdir=foo is equivalent to --outputdir=foo</li>
 * </ul>
 *
 *
 *
 */
public class GnuParser extends Parser {
    public static final Logger log = Logger.getLogger(GnuParser.class.getName());

    //according to http://www.urbanophile.com/arenn/hacking/getopt
    //POSIX correctness means treating all options following teh first arg as args
    private boolean posixlyCorrect;

    //  private String invalidArgumentPattern="(--)|(-).*";

    public GnuParser(Option... options) {
        super(options);
    }

    public GnuParser(Option[] options, Argument[] args) {
        super(options, args);
        checkOptions(Arrays.asList(options));
    }

    public GnuParser(List<Option> options, List<Argument> args) {
        super(options, args);
        checkOptions(options);
    }

    protected void checkShortOptionsUnique(List<Option> options){
        Map<Character,Option> s = new HashMap<Character,Option>(options.size(),1f);
        for(Option o:options){
            if (o.hasShortId()) {
                final Character key = new Character(o.getShortId());
                Option old;
                if((old=s.put(key,o))!=null){
                    throw new ConfigException("may only have one short option with any given shortId!" +
                            " option="+old+" and option="+o+" both have a shortId of "+key);
                }
            }
        }
    }
   

    protected void checkOptions(List<Option> options) {
        checkShortOptionsUnique(options);
        //check there are no opts with a shortId of -W
        for(Option o:options){
            if(o.hasShortId() && 'W'==o.getShortId()){
                throw new ConfigException("Cannot have a short option with an id of W as this" +
                        " id is reserved for a special POSIX long option syntax");
            }
        }
    }

    /**
     * as far as I can tell without having done much research the GNU style and POSIX style of option
     * handling are very similar; so not worth having a separate POSIX parser.. for now.
     * This flag switches to strict POSIX compliance, e.g stop parsing options after teh 1st argument is found
     * By default is off
     * @param posixlyCorrect
     */
    public void setPosixlyCorrect(boolean posixlyCorrect) {
        this.posixlyCorrect = posixlyCorrect;
    }


//     public void setInvalidArgumentPattern(String invalidArgumentPattern) {
//        this.invalidArgumentPattern = invalidArgumentPattern;
//    }


    public void parseArgs(String[] argv)
            throws
            MissingOptException, MissingArgException, InvalidArgException, MissingOptArgException, AmbiguousOptionNameException, InvalidOptArgException, UnrecognizedOptionException, CollapsedOptionsWithArgumentException, IncompleteGroupException, InvalidGroupException, InvalidOptSyntaxException{

        boolean atArgs=false;
        int argIndex=0;
        for (int i = 0; i < argv.length; i++) {
            String s = argv[i];
             if("--".equals(s)){ //everything from now on is arg
                atArgs=true;
                continue;
            }
            if(!atArgs && s.startsWith("-")){ //got 1 or more options
                i=readOptions(s,i,argv);
            }
            else{//got an arg
                if(posixlyCorrect) atArgs=true;
                readArg(argIndex, s);
                argIndex++;
            }
        }
    }



    /**
     * attempt to read an option from argv
     * @param optionId option name (e.g. -a -alpha)
     * @param argvIndex position in argv
     * @param argv commandline
     * @return index to argv following the successfully read option, or -1 if no read
     */
    private int readOptions(String optionId, int argvIndex, String[] argv)
            throws AmbiguousOptionNameException, MissingOptArgException, InvalidOptArgException, UnrecognizedOptionException, CollapsedOptionsWithArgumentException, InvalidOptSyntaxException {
        if(optionId.startsWith("-W")){
            //For example, "-W outputdir=foo" would be equivalent to "--outputdir=foo".
            //hacky: convert from first form to 2nd and continue
            if(++argvIndex >=argv.length) throw new InvalidOptSyntaxException("the -W POSIX option must have an argument (of the form x=y)");
            return readLongOption("--"+argv[argvIndex], argvIndex, argv);
        }
        else if(optionId.startsWith("--")){
            return readLongOption(optionId, argvIndex, argv);
        }
        else{
            return parseShortOption(optionId, argvIndex, argv);
        }
    }

    private int parseShortOption(String optionId, int argvIndex, String[] argv) throws MissingOptArgException, InvalidOptArgException, CollapsedOptionsWithArgumentException, UnrecognizedOptionException {
        assert optionId.startsWith("-"):"invalid shortoption:"+optionId;
        String optNames = optionId.substring(1);
        if(optNames.length()==1){ //just the one option
            char optname=optNames.charAt(0);
            boolean found=false;
            for(Option o:options){
                char id=o.getShortId();
                if(id>0 && id==optname){ //got the right option
                    log.fine("input="+optionId+" matches option="+o);
                    if(o.takesArgument()){
                        if(++argvIndex>=argv.length) throw new MissingOptArgException(o);
                        String value=argv[argvIndex];
                        try {
                            log.fine("read value="+value);
                            o.readArg(value);
                        } catch (InvalidArgException e) {
                            throw new InvalidOptArgException(o,e);
                        }
                    }
                    o.setFound(true);
                    found=true;
                    break;
                }
            }
            if(!found) throw new UnrecognizedOptionException(optionId,String.valueOf(optname));
        }
        else{//multiple options e.g. -txv
            for(char optname: optNames.toCharArray()){
                boolean found=false;
                for(Option o:options){
                    char id=o.getShortId();
                    if(id>0 && id==optname){ //got the right option
                        log.fine("input="+optionId+" matches option="+o);
                        if(o.takesArgument()){
                            throw new CollapsedOptionsWithArgumentException(o,optionId);
                        }
                        o.setFound(true);
                        found=true;
                        break;
                    }
                }
                if(!found) throw new UnrecognizedOptionException(optionId,String.valueOf(optname));
            }
        }
        return argvIndex;
    }

    private int readLongOption(String optionId, int argvIndex, String[] argv) throws MissingOptArgException, InvalidOptArgException, AmbiguousOptionNameException, UnrecognizedOptionException {
        final String longId;
        int indexOfEquals;
        if((indexOfEquals=optionId.indexOf('='))!=-1){
            longId=optionId.substring("--".length(),indexOfEquals);
        }
        else{
            longId=optionId.substring(2);
        }
        LinkedList<Option> matchesL=new LinkedList<Option>();
        for(Option o:options){
            final String _longId = o.getLongId();
            if(_longId!=null && _longId.startsWith(longId)){
                matchesL.add(o);
            }
        }
        if(matchesL.size()==1){
            Option o = matchesL.getFirst();
            log.fine("input="+optionId+": matched option="+o);
            if (o.takesArgument()) {
                final String value;
                if(indexOfEquals!=-1){ //is --abc=d style option
                    value=optionId.substring(indexOfEquals+1);
                }
                else{
                    if(++argvIndex>=argv.length) throw new MissingOptArgException(o);
                    value=argv[argvIndex];
                }
                try {
                    log.fine("read value="+value);
                    o.readArg(value);
                } catch (InvalidArgException e) {
                    throw new InvalidOptArgException(o,e);
                }
            }
            o.setFound(true);
        }
        else if(matchesL.size()>1){
            throw new AmbiguousOptionNameException(optionId,matchesL);
        }
        else throw new UnrecognizedOptionException(longId);
        return argvIndex;
    }

    //todo: this usage style is more like jakarta CLI than GNU.. make a CLI-style parser with this usage code
    public String getUsage(String usageHeader, String usageFooter){
        StringBuilder sb = new StringBuilder();
        sb.append(usageHeader);

        if(!options.isEmpty()){
            sb.append("\n");
            sb.append(getOptionsString(1,80)); //1 char for teh newline character
            boolean allOptionsOptional=true;
            if(!options.isEmpty()){
                for(Option o:options) {
                    if(o.isRequired()) {
                        allOptionsOptional=false;
                        break;
                    }
                }
            }
            if (!allOptionsOptional) {
                sb.append("* required options\n");
            }
        }
        sb.append("\n");
        sb.append(usageFooter);
        return sb.toString();
    }


}
