package uk.co.hcsoftware.yago.util;

import uk.co.hcsoftware.yago.ArgReader;
import uk.co.hcsoftware.yago.InvalidArgException;
import uk.co.hcsoftware.yago.ConfigException;

import java.util.logging.Logger;
import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.File;
import java.text.*;
import java.text.ParseException;


public class DefaultArgReaders {
    public static final Logger log = Logger.getLogger(DefaultArgReaders.class.getName());
    public static ArgReader<String> createStringReader(){
        return new ArgReader<String>() {
            public String readArg(String argValue){
                return argValue;
            }
        };
    }
    public static ArgReader<Integer> createIntReader(){
        return new ArgReader<Integer>() {
            public Integer readArg(String argValue) throws InvalidArgException {
                try {
                    final int i = Integer.parseInt(argValue);
                    return new Integer(i);
                } catch (NumberFormatException e) {
                    throw new InvalidArgException("cannot convert to integer",e,argValue);
                }
            }
        };
    }
    public static ArgReader<File> createFileReader(final boolean checkExists){
        return new ArgReader<File>() {
            public File readArg(String argValue) throws InvalidArgException {
                File f = new File(argValue);
                if(checkExists && !f.exists()){
                    throw new InvalidArgException("f="+f+" doesn't exist",argValue);
                }
                return f;
            }
        };
    }
    /**
     *  parsers will be called in order of formats, the first to successfully parse the date will return it
     *
     * @param simpleDateFormats list of formats that will be passed to constructor of java.text.SimpleDateFormat
     */
    public static ArgReader<Date> createDateReader(List<String> simpleDateFormats){
        if(simpleDateFormats.size()<1) throw new ConfigException("must be >1 dateformats");
        final List<DateFormat> dfs = new ArrayList<DateFormat>(simpleDateFormats.size());
        for(String s:simpleDateFormats) dfs.add(new SimpleDateFormat(s));
        return new ArgReader<Date>() {
            public Date readArg(String argValue) throws InvalidArgException {
                for(DateFormat df:dfs){
                    try {
                        return df.parse(argValue);
                    } catch (ParseException e) {
                        throw new InvalidArgException("cannot parse date",e,argValue);
                    }
                }
                log.severe("shouldn't get here!");
                throw new RuntimeException("internal error!"){};
            }
        };
    }
    public static ArgReader<Date> createDateReader(String ... simpleDateFormats){
        return createDateReader(Arrays.asList(simpleDateFormats));
    }

}
